from statistics import median_high
from . import AL
import timeit
import numpy as np
'''
Run from parent folder with:
python3 -m avispa_lattices.test
'''


def test_M3():
    L = AL.Lattice.from_up_edges(5, [(0, 1), (0, 2), (0, 3), (1, 4), (2, 4),
                                     (3, 4)])
    assert L.is_poset
    assert L.is_lattice
    assert L.is_modular
    assert not L.is_distributive, L.show()
    return


def test_M5():
    L = AL.Lattice.from_up_edges(5, [(0, 1), (1, 2), (0, 3), (2, 4), (3, 4)])
    assert L.is_poset
    assert L.is_lattice
    assert not L.is_modular
    assert not L.is_distributive
    return


def test_generation_until_7():
    non_mod = []
    mod = []
    dist = []
    for L in AL.all_lattices(7):
        if L.is_distributive:
            dist.append(L)
        elif L.is_modular:
            mod.append(L)
        else:
            non_mod.append(L)

    assert len(dist) == 22
    assert len(mod) == 12
    assert len(non_mod) == 45


def test_generation_and_f_iteration():
    R = np.random.RandomState(12382332)
    seeds = R.randint(0, 2**32, size=100)
    lat_list = [AL.random_lattice(20, seed=s, method='Czech') for s in seeds]
    for L in lat_list:
        assert L.is_poset
        assert L.is_lattice
        pass
    return


L = AL.random_lattice(6, seed=42)
#help(L.f_iter)
L.show()
# for f in L.f_iter('all', n=3):
#     print(f)
# for f in L.f_iter('monotones', n=3):
#     print(f)
#     L.show(f)
F = [f for f in L.f_iter('lub', n=30)]
print(F[0], F[10], F[20], sep='\n')
AL.graphviz.COLORS_CYCLE = ['darkblue', 'darkgreen', 'darkorange', 'darkred']
L.show(F[0], F[10], L.f_lub(F[0], F[10]))
L.show(F[0], F[10], L.f_lub(F[0], F[10]), L.f_glb(F[0], F[10]))
L.show(F[0], F[10], L.f_lub(F[0], F[10]),
       L.f_glb(F[0], F[10], method='pointwise'))


def test_f_glb():
    R = np.random.RandomState(123976)
    #P = AL.random_poset
    L = AL.random_lattice(5, seed=R.randint(0, 2**32), method='Czech')
    L.show()
    print(L.is_distributive)
    # for f in L.f_iter(method='monotones'):
    #     print(f)
    for f in L.f_iter():
        print(f)
        print('HEY')
    return


# test_f_glb()
# test_generation_and_f_iteration()
# test_M3()
# test_M5()
# test_generation_until_7()
#assert 2 < timeit.timeit(lambda: test_generation_until_7(), number=3) < 5



from __future__ import annotations
from functools import reduce
from optparse import Option
from typing import TYPE_CHECKING, Any, Iterable, Optional, Tuple, cast, List, Sequence
from typing_extensions import Literal, get_args as literal_args
from ._types import PartialEndomorphism, Endomorphism

if TYPE_CHECKING:
    from .. import Poset, Lattice

from ..utils.iterators import product_list
from itertools import islice

_f_iter_method = Literal['auto', 'lub', 'all', 'monotones', 'lub_no_bottom']
_f_iter_methods: Tuple[str] = literal_args(_f_iter_method)


def f_iter(L: Lattice, method: _f_iter_method = 'auto', n: Optional[int] = None,
           in_place=False):
    '''
    Iterator of endomorphisms over a given lattice.
    method can be:
        'all': all endomorphisms
        'monotones': all monotone endomorphisms
        'lub': all join-endomorphisms, i.e. endomorphisms that preserve least-upper-bounds.
        'lub_no_bottom': all endomorphisms that preserve lub but without the constraint f[bottom]=bottom.
        'auto': (default) defaults to "lub"

    if in_place is True (faster, but advanced), the iterator will succesively yield the same endomorphism object with entries modified
    '''
    if method == 'auto':
        method = 'lub'
    return islice((cast(Endomorphism, f.copy()) for f in _f_iter(L, method)), n)


def _f_iter(L: Lattice, method: _f_iter_method):
    if method == 'lub' and L.is_distributive:
        yield from f_iter_lub_distributive(L)
    elif method == 'lub':
        for f in f_iter_monotone(L):
            if f_is_lub(L, f):
                yield f
    elif method == 'all':
        yield from f_iter_all(L)
    elif method == 'monotones':
        yield from f_iter_monotone_no_bottom(L)
    elif method == 'lub_no_bottom':
        for f in f_iter_monotone_no_bottom(L):
            if f_is_lub_no_bottom(L, f):
                yield f
    else:
        raise NotImplementedError(f'"{method}" not in {_f_iter_methods}')


def f_iter_lub_distributive(self: Lattice):
    'all functions that preserve lubs for sets'
    yield from f_iter_irreducibles_monotone(self)


def f_iter_all(self: Poset):
    'all endomorphisms'
    return product_list(range(self.n), repeat=self.n)


def num_f_all(self: Poset):
    return self.n**self.n


def f_iter_all_bottom(self: Lattice):
    'all endomorphisms f with f[bottom]=bottom'
    n = self.n
    if n > 0:
        options = [range(n) if i != self.bottom else [i] for i in range(n)]
        for f in product_list(*options):
            yield f
    return


def num_f_all_bottom(self: Poset):
    return self.n**(self.n - 1)


def f_is_monotone(self: Poset, f, domain=None):
    'check if f is monotone over domain'
    n = self.n
    domain = range(n) if domain is None else domain
    leq = self.leq
    for i in domain:
        for j in domain:
            if leq[i, j] and not leq[f[i], f[j]]:
                return False
    return True


def f_iter_monotone_bruteforce(self: Poset):
    'all monotone functions'
    for f in f_iter_all(self):
        if f_is_monotone(self, f):
            yield f
    return


def f_iter_monotone_bottom_bruteforce(self: Lattice):
    'all monotone functions with f[bottom]=bottom'
    for f in f_iter_all_bottom(self,):
        if f_is_monotone(self, f):
            yield f
    return


def f_iter_monotone_no_bottom(self: Lattice):
    'all monotone functions'
    f: PartialEndomorphism = [None] * self.n
    yield from f_iter_monotone_restricted(self, _f=f)


def f_iter_lub_bruteforce(self: Lattice):
    'all space functions. Throws if no bottom'
    for f in f_iter_monotone(self,):
        if f_is_lub_no_bottom(self, f):
            yield f
    return


def f_iter_monotone_restricted(self: Lattice, domain=None,
                               _f: Optional[PartialEndomorphism] = None):
    'generate all monotone functions f : domain -> self, padding non-domain with None'
    n = self.n
    leq = self.leq
    geq_list = [[j for j in range(n) if leq[i, j]] for i in range(n)]
    f: PartialEndomorphism = [None for i in range(n)] if _f is None else _f
    topo, children = self.MD.graphs._toposort_children(self, domain)
    yield from _f_iter_monotone_restricted(self, f, topo, children, geq_list)


def _f_iter_monotone_restricted(
    self: Lattice,
    f: PartialEndomorphism,
    topo: List[int],
    children,
    geq_list: List[List[int]],
):
    m = len(topo)
    lub_f = self.lub_of_many

    def backtrack(i):
        'f[topo[j]] is fixed for all j<i. Backtrack f[topo[k]] for all k>=i, k<m'
        if i == m:
            yield f
        else:
            for k in geq_list[lub_f(children[i])]:
                f[topo[i]] = k  # type:ignore
                yield from backtrack(i + 1)

    yield from backtrack(0)


def f_iter_monotone(self: Lattice):
    'all monotone functions with f[bottom]=bottom'
    if self.n == 0:
        return
    f = cast(List[int], [None] * self.n)
    f[self.bottom] = self.bottom
    domain = [i for i in range(self.n) if i != self.bottom]
    yield from f_iter_monotone_restricted(self, domain=domain, _f=f)


def _interpolate_funcs(self: Lattice, funcs, domain) -> Iterable[List[int]]:
    'extend each f in funcs outside domain using f[j]=lub(f[i] if i<=j and i in domain)'
    n = self.n
    lub = self.lub
    leq = self.leq
    bot = self.bottom
    no_domain = [i for i in range(n) if i not in domain]
    dom_leq = [[i for i in domain if leq[i, j]] for j in range(n)]
    lub_f = (lambda a, b: lub[a, b])
    for f in funcs:
        for j in no_domain:
            f[j] = reduce(lub_f, (f[x] for x in dom_leq[j]), bot)
        yield f


def f_iter_irreducibles_monotone(self: Lattice) -> Iterable[List[int]]:
    'all functions given by f[non_irr]=lub(f[irreducibles] below non_irr)'
    if self.n == 0:
        return
    n = self.n
    leq = self.leq
    geq_list = [[j for j in range(n) if leq[i, j]] for i in range(n)]
    m, m_topo, m_children = self.irreducible_components
    f = [None for i in range(n)]
    f = cast(PartialEndomorphism, f)

    def backtrack(i):
        if i == m:
            yield f
        else:
            it = _f_iter_monotone_restricted(
                self,
                f,
                m_topo[i],
                m_children[i],
                geq_list,
            )
            for _ in it:
                yield from backtrack(i + 1)

    funcs = backtrack(0)
    yield from _interpolate_funcs(self, funcs, self.irreducibles)


def f_iter_irreducibles_monotone_no_bottom(self: Lattice):
    'all functions given by f[non_irr]=lub(f[irreducibles] below non_irr) and'
    'f[bottom] = any below or equal to glb(f[irreducibles])'
    n = self.n
    if n == 0:
        return
    glb = self.glb
    leq = self.leq
    below = [[i for i in range(n) if leq[i, j]] for j in range(n)]
    bottom = self.bottom
    irreducibles = self.irreducibles
    for f in f_iter_irreducibles_monotone(self,):
        _glb_f = (lambda acum, b: glb[acum, f[b]])
        glb_f = lambda elems: reduce(_glb_f, elems, self.top)
        for i in below[glb_f(irreducibles)]:
            f[bottom] = i
            yield f


'''
@section
    Methods for endomorphisms that preserve lub
'''


def f_is_lub(self: Lattice, f, domain=None):
    '''
    check if f preserves lubs for sets:
        f_is_lub_no_bottom and f[bottom]=bottom.
    Throws if no bottom
    '''
    n = self.n
    # Small outlier cases
    if n == 0 or (domain is not None and len(domain) <= 1):
        return True
    # Check bottom
    bot = self.bottom
    if f[bot] != bot or (domain is not None and bot not in domain):
        return False
    # Check the rest
    return f_is_lub_no_bottom(self, f, domain)


def f_is_lub_no_bottom(L: Lattice, f, domain=None):
    '''
    check if f preserves lubs for all pairs:
        f[lub[i,j]]=lub[f[i],f[j]]
    '''
    n = L.n
    lub = L.lub
    domain = range(n) if domain is None else domain
    for i in domain:
        for j in domain:
            if f[lub[i, j]] != lub[f[i], f[j]]:
                return False
    return True


def f_iter_lub_no_bottom_bruteforce(self: Lattice):
    'all functions that statisfy f_is_lub_no_bottom'
    for f in f_iter_monotone_no_bottom(self,):
        if f_is_lub_no_bottom(self, f):
            yield f
    return


def f_iter_lub_no_bottom(self: Lattice):
    'all functions that statisfy f_is_lub'
    it = f_iter_irreducibles_monotone_no_bottom(self,)
    if self.is_distributive:
        yield from it
    else:
        for f in it:
            if f_is_lub_no_bottom(self, f):
                yield f


def num_f_lub_no_bottom(self: Lattice):
    return count_f_lub_no_bottom_bruteforce(self,)


def count_f_lub_no_bottom_bruteforce(self: Lattice):
    return sum(1 for f in f_iter_lub_no_bottom(self,))


# def num_f_lub(self: Lattice):
#     return count_f_lub(self,)

# def count_f_lub(self: Lattice):
#     if self.is_distributive:
#         num = count_f_lub_distributive(self)
#     else:
#         num = count_f_lub_non_distributive(self)
#     return num

# def count_f_lub_non_distributive(self: Lattice):
#     return sum(1 for f in f_iter_lub(self,))


def f_is_lub_of_irreducibles(self: Lattice, f, domain=None):
    '''
    check if f satisfies for all x in range(n) that
        f(x) == self.set_lub(*[f[i] for i in I(x)])
    where
        I(x) = list of irreducibles below (leq) x
    '''
    n = self.n
    set_lub = self.lub_of_many
    I = self.irreducible_descendants
    for a in range(n):
        if f[a] != set_lub(I[a]):
            return False
    return True


def count_f_lub_distributive(L: Lattice):
    '''
    This function assumes that your lattice is distributive
    '''
    if L.n == 0:
        return 0
    n = L.n
    leq = L.leq
    geq_list = [[j for j in range(n) if leq[i, j]] for i in range(n)]
    m, m_topo, m_children = L.irreducible_components
    f = [None for _ in range(n)]

    f = cast(PartialEndomorphism, f)

    def num(i: int):
        'num of monotone functions restricted to domain k_topo[i]'
        it = _f_iter_monotone_restricted(
            L,
            f,
            m_topo[i],
            m_children[i],
            geq_list,
        )
        return sum(1 for _ in it)

    k_independent = [num(k) for k in range(m)]
    return reduce(lambda a, b: a * b, k_independent, 1)


from __future__ import annotations
from functools import lru_cache
from typing import TYPE_CHECKING, List, Optional, Sequence, Union, cast
if TYPE_CHECKING:
    from ..base import Lattice, Poset, Relation

import numpy as np
from ..utils.numpy_types import npBoolMatrix, npUInt64Matrix
from collections import deque


def floyd_warshall(adj: npBoolMatrix, infinity: int) -> npUInt64Matrix:
    'Compute all pairs shortest distances using Floyd-Warshall algorithm'
    dist: npUInt64Matrix
    dist = adj.astype(np.uint64)  # type:ignore
    dist[~adj] = infinity
    dist[np.diag_indices_from(dist)] = 0
    for k in range(len(dist)):
        np.minimum(dist, dist[:, k, None] + dist[None, k, :], out=dist)
    return dist


def heights(self: Poset):
    'Array of distance from i down to any bottom'
    dist = self.dist
    bottoms = self.bottoms
    return list(np.min([dist[i, :] for i in bottoms], axis=0))


def depths(self: Poset):
    'Array of distance from any top down to i'
    dist = self.dist
    tops = self.tops
    return list(np.min([dist[:, i] for i in tops], axis=0))


def height(self: Poset):
    return max(heights(self))


def _parse_domain(n: int, domain: Sequence[int] | Sequence[bool]) -> List[int]:
    assert len(domain) <= n, f'Invalid domain: {domain}'
    if len(domain) == n > 0:
        if isinstance(domain[0], bool):
            domain = [i for i in range(n) if domain[i]]
    else:
        assert len(set(domain)) == len(domain), f'Invalid domain: {domain}'
    return domain  # type:ignore


def subgraph(P: Poset, domain: Sequence[int] | Sequence[bool]):
    domain = _parse_domain(P.n, domain)
    m = len(domain)
    leq = P.leq
    sub = np.zeros((m, m), dtype=bool)
    for i in range(m):
        for j in range(m):
            sub[i, j] = leq[domain[i], domain[j]]
    sub.flags.writeable = False
    P_labels = P._labels
    labels = tuple(P_labels[i] for i in domain)
    return P.__class__(sub, check=False, labels=labels)


def toposort_bottom_up(P: Poset):
    n = P.n
    G = P.parents
    child = P.child
    indeg = [child[:, i].sum() for i in range(n)]
    topo: List[int] = []
    q = deque([i for i in range(n) if indeg[i] == 0])
    while q:
        u = q.popleft()
        topo.append(u)
        for v in G[u]:
            indeg[v] -= 1
            if indeg[v] == 0:
                q.append(v)
    assert len(topo) == n, f'Not antisymmetric, cycle found'
    return topo


def toporank(self: Poset):
    return tuple(inverse_permutation(self.toposort_bottom_up))


def inverse_permutation(perm: Sequence[int], check=False):
    n = len(perm)
    if check:
        assert set(perm) == set(range(n)), f'Invalid permutation {perm}'
    rank = [-1] * n
    for i in range(n):
        rank[perm[i]] = i
    return rank


def independent_components(self: Poset):
    'Graph components if all edges were bidirectional'
    n = self.n
    cmp = self.leq | self.leq.T
    G = [[j for j in range(n) if cmp[i, j]] for i in range(n)]
    color = np.ones(n, dtype=int) * -1

    def component(i: int):
        q = deque([i])
        found = []
        while q:
            u = q.popleft()
            for v in G[u]:
                if color[v] != color[u]:
                    color[v] = color[u]
                    q.append(v)
            found.append(u)
        return found

    comps: List[List[int]] = []
    for i in range(n):
        if color[i] == -1:
            color[i] = len(comps)
            comps.append(component(i))
    return comps


def transitive_reduction(leq: npBoolMatrix):
    ''''
    Compute in O(n^3) the transitive reduction of the given relation
    Assuming that it is a poset
    The output relation is also known as "Hasse diagram"
    '''
    lt = leq.copy()
    lt[np.diag_indices_from(lt)] = False
    any_inbetween = np.matmul(lt, lt)
    child = lt & ~any_inbetween
    return cast(npBoolMatrix, child)


def bottoms(self: Poset):
    'bottom elements of the poset'
    n = self.n
    nleq = self.leq.sum(axis=0)
    return [i for i in range(n) if nleq[i] == 1]


def non_bottoms(self: Poset):
    'non-bottom elements of the poset'
    n = self.n
    nleq = self.leq.sum(axis=0)
    return [i for i in range(n) if nleq[i] > 1]


def tops(self: Poset):
    'top elements of the poset'
    n = self.n
    nleq = self.leq.sum(axis=0)
    return [i for i in range(n) if nleq[i] == n]


def non_tops(self: Poset):
    'non-top elements of the poset'
    n = self.n
    nleq = self.leq.sum(axis=0)
    return [i for i in range(n) if nleq[i] < n]


def _toposort_children(self: Poset, domain: Optional[Sequence[int]]):
    'Compute a toposort for domain and the children lists filtered for domain'
    'j in out.children[i] iff j in out.topo and j is children of out.topo[i]'
    n = self.n
    D = range(n) if domain is None else domain
    topo = [i for i in self.toposort_bottom_up if i in D]
    sub = self.MD.graphs.subgraph(self, topo)
    children = [[topo[j] for j in l] for l in sub.children]
    return topo, children


from typing import Optional
from typing_extensions import Literal
from packaging.version import parse

VERSION = '3.0.9'


def version_cmp(a: str, b: Optional[str] = None) -> Literal[0, -1, 1]:
    if b is None:
        b = VERSION
    A = parse(a)
    B = parse(b)
    if A == B:
        return 0
    if A < B:
        return -1
    return 1


def version_is_at_least(a: str):
    return version_cmp(a) <= 0



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..base import Lattice, Poset, Relation
from typing import Sequence, Tuple


def describe(P: Poset):
    print('Relation matrix:')
    print(P.leq.astype(int))
    print('Covers:', P)
    print(f'Lattice? {P.is_lattice}')
    if P.is_lattice:
        pass
        # print(f'Distributive? {P.is_distributive}')
    # else:
    #     print(f'# bottoms: {len(bottoms(P))}')
    #     print(f'# tops: {len(tops(P))}')
    P.show()
    return


def name(self: Poset):
    'Compact and readable representation of self based on parents'
    if not hasattr(self, 'n'):
        return f'Uninitialized {self.__class__.__name__}'
    n = self.n
    P = self.parents
    topo = self.toposort_bottom_up
    Pstr = lambda i: ','.join(map(str, P[i]))
    it = (f'{i}<-{Pstr(i)}' for i in topo if P[i])
    name = ' : '.join([f'{n}', *it])
    labels = ''
    if self.labels != tuple(range(n)):
        labels = ', '.join(self._labels)
        labels = f' with labels {labels}'
    return f'P({name}){labels}'


from ._version import VERSION

# Classes and functions
from .base import Lattice, Poset, Relation
from .base_generation.exhaustive import all_lattices
from .base_generation.random import lattice as random_lattice
from .base_generation.random import poset as random_poset
from .functional import f_lub, f_glb, f_iter

# Modules
from . import functional
from .base_methods import validation, graphviz
from .base_generation import random, exhaustive

# from typing import TYPE_CHECKING, Iterable, Callable, List, Optional, Union, cast
# if TYPE_CHECKING:
#     from .base import Endomorphism



import re
import numpy as np

npBoolMatrix = np.ndarray
npUInt64Matrix = np.ndarray
npInt64Array = np.ndarray

_get_dtype_string = re.compile(
    r'(<class \'numpy\.(.*)\'>)|(<class \'(.*?)\'>)|(.*)')


def get_dtype_string(dtype):
    'return the dtype string of a numpy dtype'
    m = _get_dtype_string.match(str(dtype))
    assert m
    g = m.groups()
    dtype_str: str = g[1] or g[3] or g[4]
    np_dtype = np.dtype(dtype_str)  # type:ignore
    assert dtype == np_dtype, (
        f'Non invertible dtype: {dtype} != np.dtype(\'{dtype_str}\')')
    return dtype_str



from __future__ import annotations
from collections import deque
from typing import TYPE_CHECKING, Optional

from ..base import Lattice

import numpy as np
from itertools import chain
from ..utils.iterators import cartesian


def _add_edge(self: Lattice, i, j, assume_poset=False):
    "Grow self by adding one edge 'i leq j'"
    leq = self.leq
    new_leq = leq + np.matmul(leq[:, i:i + 1], leq[j:j + 1, :])
    new_leq.flags.writeable = False
    obj = self.__class__(new_leq, check=False)
    if not assume_poset:
        _ = obj.toposort_bottom_up
    return obj


def _add_node(self: Lattice, i, j, assume_poset=False):
    "Grow self by adding one node just between i and j"
    n = self.n
    leq = self.leq
    out = np.zeros((n + 1, n + 1), bool)
    out[:-1, :-1] = leq
    out[n, n] = True
    out[:-1, :-1] += np.matmul(leq[:, i:i + 1], leq[j:j + 1, :])
    out[n, :-1] = leq[j, :]
    out[:-1, n] = leq[:, i]
    out.flags.writeable = False
    obj = self.__class__(out, check=not assume_poset)
    return obj


def forbidden_pairs(self: Lattice):
    "Pairs (i,j) that break lub uniqueness or partial order structure"
    n = self.n
    leq = self.leq
    joi = self.lub
    nocmp = ~(leq + leq.T)

    def f(a, b):
        if leq[b, a]:
            return True
        if leq[a, b]:
            return False
        X = [x for x in range(n) if leq[x, a]]
        Y = [y for y in range(n) if ~leq[b, y] and nocmp[y, a]]
        return any(nocmp[joi[x, y], joi[b, y]] for y in Y for x in X)

    fb = np.array([[f(i, j) for j in range(n)] for i in range(n)], dtype=bool)
    return fb


def iter_add_edge(self: Lattice):
    "Grow self by adding one edge"
    n = self.n
    leq = self.leq
    fb = forbidden_pairs(self)
    vis = set()
    h = self.hash_elems
    for i, j in cartesian(n, n):
        if not fb[i, j] and not leq[i, j] and (h[i], h[j]) not in vis:
            yield _add_edge(self, i, j, assume_poset=True)
    return


def iter_add_node(self: Lattice):
    "Grow self by adding one node"
    n = self.n
    fb = forbidden_pairs(self)
    vis = set()  # Don't repeat isomorphical connections
    h = self.hash_elems
    for i, j in cartesian(n, n):
        if not fb[i, j] and not (h[i], h[j]) in vis:
            yield _add_node(self, i, j, assume_poset=True)
    return


def iter_all_latices(max_size: int, starting_lattice: Optional[Lattice] = None):
    q: deque[Lattice]
    if starting_lattice is None:
        q = deque([Lattice.from_children(x) for x in [[], [[]], [[], [0]]]])
    else:
        q = deque([starting_lattice])
    vis = set()
    while q:
        U = q.popleft()
        yield U.canonical
        it = iter_add_node(U) if U.n < max_size else iter([])
        for V in chain(iter_add_edge(U), it):
            if V not in vis:
                vis.add(V)
                q.append(V)
    return



from __future__ import annotations
import functools
import itertools
from typing import TYPE_CHECKING, Callable, Iterable, List, Optional, Sequence, Tuple, Union, cast
from typing_extensions import Literal, get_args as literal_args
from collections import deque
from .._version import version_is_at_least
if TYPE_CHECKING:
    from ..base import Lattice, Poset, Relation
    from ._types import Endomorphism, PartialEndomorphism


def f_lub(self, *functions: Endomorphism) -> Endomorphism:
    'Pointwise lowest upper bound of a set of functions'
    return f_lub_pointwise(self, functions)


_f_glb_method = Literal['auto', 'pointwise', 'GMeet', 'GMeet+', 'DMeet+',
                        'JMeet', 'CMeet',]
_f_glb_methods: Tuple[str] = literal_args(_f_glb_method)


def f_glb(L: Lattice, *functions: Endomorphism,
          method: _f_glb_method = 'auto') -> Endomorphism:
    '''
    Greatest lower bound of a set of functions.
    '''
    if method == 'auto':
        if L.is_distributive:
            method = 'DMeet+'
        else:
            if version_is_at_least('3.1.0'):
                method = 'GMeet+'
            else:
                method = 'GMeet'
    if method == 'pointwise':
        return f_glb_pointwise(L, functions)
    elif method == 'GMeet':
        return f_glb_GMeet(L, functions)
    elif method == 'GMeet+':
        raise NotImplementedError
    elif method == 'DMeet+':
        return f_glb_DMeet_plus(L, functions)
    else:
        raise NotImplementedError(f'"{method}" not in {_f_glb_methods}')


def f_glb_pointwise(L: Lattice,
                    functions: Iterable[Endomorphism]) -> Endomorphism:
    '''
    Pointwise greatest lower bound of a set of functions.

    This implementation avoids computing the default every
    time and supports comprehensions as input.
    '''
    try:
        first, *functions = functions
    except ValueError:
        h = [L.top] * L.n
        h[L.bottom] = L.bottom
        return h
    h = functools.reduce(
        lambda f, g: [L.glb[f[i], g[i]] for i in range(L.n)],
        functions,
        first,
    )
    return h


def f_lub_pointwise(L: Lattice,
                    functions: Iterable[Endomorphism]) -> Endomorphism:
    '''
    Pointwise lowest upper bound of a set of functions.

    This implementation avoids computing the default every
    time and supports comprehensions as input.
    '''
    try:
        first, *functions = functions
    except ValueError:
        return [L.bottom] * L.n
    h = functools.reduce(
        lambda f, g: [L.lub[f[i], g[i]] for i in range(L.n)],
        functions,
        first,
    )
    return h


def f_glb_GMeet(L: Lattice, functions: Iterable[Endomorphism]) -> Endomorphism:
    """
    Greatest lower bound of a set of lub-functions.

    Algorithm 2 in "Counting and Computing Join-Endomorphisms in Lattices (Revisited)*"
    """
    h = f_glb_pointwise(L, functions)
    h = fix_f_naive(L, h)
    return h


def fix_f_naive(L: Lattice, f: Endomorphism,
                budget: Optional[int] = None) -> Endomorphism:
    '''
    Compute the "first" function h below f that is a space function
    by fixing iteratively the pairs of elements in the lattice that
    fail to satisfy the LUB axiom

    ... The complexity is, I think, O(n^4). Needs to be checked.
    '''
    lub = L.lub
    glb = L.glb
    leq = L.leq
    it = itertools.count() if budget is None else range(budget)
    f_prev = f = f.copy()
    for _ in it:
        for i in range(L.n):
            for j in range(L.n):
                k = lub[i, j]
                fi_lub_fj = lub[f[i], f[j]]
                if leq[f[k], fi_lub_fj]:
                    f[k] = fi_lub_fj
                elif f[k] != fi_lub_fj:
                    f[i] = glb[f[i], f[k]]
                    f[j] = glb[f[j], f[k]]
        if f == f_prev:
            break
        f_prev = f.copy()
    return f


def f_glb_DMeet_plus(L: Lattice,
                     functions: Sequence[Endomorphism]) -> Endomorphism:
    """
    Greatest lower bound of a set of lub-functions in O(n*m).
        n = L.n
        m = len(functions)
    
    This implementation takes advantage of the join irreducible
    elements to reduce the number of operations.
    It is explained in "Counting and Computing Join-Endomorphisms in Lattices (Revisited)*"

    Adapted from the implementation delta_plus_jies at:
        https://github.com/Sirquini/delta/blob/master/delta.py
    """
    n = L.n
    if len(functions) <= 1:
        return f_glb_pointwise(L, functions)
    covers = L.children

    h: PartialEndomorphism = [None for _ in range(n)]
    h[L.bottom] = L.bottom
    for j in L.irreducibles:
        h[j] = L.glb_of_many(fn[j] for fn in functions)

    work = deque(range(n))
    while work:
        x = work.popleft()
        if h[x] is not None:
            continue
        try:
            i, j, *_ = covers[x]
        except ValueError as e:
            raise e from Exception(
                'Invariant violated: "If x has at most 1 cover then h[x] != None"'
            )
        if h[i] is None or h[j] is None:
            if h[i] is None:
                work.append(i)
            if h[j] is None:
                work.append(j)
            work.append(x)
        else:
            h[x] = L.lub[h[i], h[j]]
    return h  # type:ignore


# def _as_external_lattice(self: Lattice):
#     '''
#     returns the equivalent "Lattice" object for self
#     to be able to run methods from external_methods
#     '''
#     mat = external_methods.lattice_from_covers(self.children)
#     return external_methods.Lattice(mat)

# def f_meet(self: Lattice, f1: Sequence[int], f2: Sequence[int]):
#     '''
#     Compute the greatest lower bound of two functions
#     in the space of join endomorphisms.
#     '''
#     return external_methods.delta_foo_cvrs(
#         _as_external_lattice(self),
#         [f1, f2],
#         self.children,
#     )


from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple
from typing_extensions import Literal, get_args as literal_args
import numpy as np
from .._version import version_is_at_least
from ..utils.algorithm_floyd_warshall import transitive_closure
from ..utils import algorithm_random_poset_czech
from ..utils.random_state import random_state
from ..base import Lattice, Poset
from ..functional._types import Endomorphism
from ..functional.operations import fix_f_naive

_random_poset_method = Literal['auto', 'p_threshold']
_random_poset_methods: Tuple[str] = literal_args(_random_poset_method)


def random_poset(n: int, seed: Optional[int] = None,
                 method: _random_poset_method = 'auto', **kwargs):
    if method == 'auto':
        method = 'p_threshold'

    if method == 'p_threshold':
        return random_poset_p(n, seed=seed, **kwargs)
    raise NotImplementedError(f'{method} not in {_random_lattice_methods}')


def random_poset_p(n: int, p: float, seed: Optional[int] = None):
    '''
    Generates a random poset.
    All posets (modulo labels) have positive probability of being generated.
    If p is close to 0, the poset is very sparse.
    If p is close to 1, the poset is very dense.
    '''
    R = random_state(seed)
    leq = np.zeros((n, n), dtype=bool)
    for i in range(n):
        for j in range(i + 1, n):
            if R.random() < p:
                leq[i, j] = 1
    for i in range(n):
        leq[i, i] = 1
    leq = transitive_closure(leq)
    leq.flags.writeable = False
    return Poset(leq, check=True)


def random_lattice_czech(n: int, seed: Optional[int] = None):
    '''
    Description: http://ka.karlin.mff.cuni.cz/jezek/093/random.pdf
    '''
    lub = algorithm_random_poset_czech.random_lattice(n, seed)
    child = (lub <= np.arange(n)[None, :])
    leq = transitive_closure(child)
    leq.flags.writeable = False
    L = Lattice(leq, check=False)
    if len(L.bottoms) >= 2:  # Collapse bottoms
        bot = L.bottoms[0]
        leq.flags.writeable = True
        leq[bot, :] = True
        leq.flags.writeable = False
        L = Lattice(leq, check=False)
    return L


_random_lattice_method = Literal['auto', 'Czech']
_random_lattice_methods: Tuple[str] = literal_args(_random_lattice_method)


def random_lattice(n: int, seed: Optional[int] = None,
                   method: _random_lattice_method = 'auto', **kwargs):
    if method == 'auto':
        if version_is_at_least('3.0.6'):
            method = 'Czech'
        else:
            method = 'Czech'
    if method == 'Czech':
        return random_lattice_czech(n, seed, **kwargs)
    raise NotImplementedError(f'{method} not in {_random_lattice_methods}')


def random_f_monotone_A(L: Lattice, seed=None,
                        _max_arbitrary_fi: Optional[int] = None):
    R = random_state(seed)
    f: Endomorphism = [-1] * L.n
    n_above = L.leq.sum(axis=1)
    p_above = n_above / n_above.sum()
    elems = np.arange(L.n)
    f[:] = R.choice(elems, L.n, p=p_above)
    if _max_arbitrary_fi is not None and _max_arbitrary_fi < L.n:
        for i in R.choice(elems, L.n - _max_arbitrary_fi):
            f[i] = L.bottom
    f[L.bottom] = L.bottom
    for i in L.toposort_bottom_up:
        min_fi = L.lub_of_many(f[k] for k in L.children[i])
        f[i] = L.lub[f[i], min_fi]
    return f


def random_f_monotone_B(L: Lattice, seed=None):
    R = random_state(seed)
    f: Endomorphism = [-1] * L.n
    n_above = L.leq.sum(axis=1)
    p_above = n_above / n_above.sum()
    elems = np.arange(L.n)
    for i in L.irreducibles:
        f[i] = R.choice(elems, 1, p=p_above)[0]
    f[L.bottom] = L.bottom
    for i in L.toposort_bottom_up:
        min_fi = L.lub_of_many(f[k] for k in L.children[i])
        assert all(f[k] != -1 for k in L.children[i])
        if f[i] != -1:
            f[i] = L.lub[f[i], min_fi]
        else:
            f[i] = min_fi
    assert L.f_is_monotone(f)
    return f


def random_f_preserving_lub(L: Lattice, seed: Optional[int] = None,
                            complexity_budget=5):
    f = _random_f_preserving_lub(L, seed=seed)
    if complexity_budget == 0:
        return f
    # Prefer complex functions
    R = random_state(seed)
    seeds = R.randint(0, 2**32, complexity_budget)
    for seed in seeds:
        new_f = _random_f_preserving_lub(L, seed=seed)
        if len(set(f)) < len(set(new_f)):
            f = new_f
    return f


def _random_f_preserving_lub(L: Lattice, seed: Optional[int] = None):
    R = random_state(seed)
    n_above = L.leq.sum(axis=1)
    p_above = n_above / n_above.sum()
    elems = np.arange(L.n)
    f: Endomorphism = [L.bottom] * L.n
    I = np.array(L.irreducibles)
    for i, fi in zip(I, R.choice(elems, len(I), p=p_above)):
        f[i] = fi
    return fix_f_naive(L, f)



from typing import List
import sys

sys.setrecursionlimit(10**6)


class Tarjan(list):
    vis: List[int]
    low: List[int]
    stack: List[int]
    in_stack: List[int]
    sccs: List[List[int]]
    vis_cnt: int

    def tarjan(self):
        G = self
        n = len(G)
        G.vis, G.low = [-1] * n, [0] * n
        G.stack, G.in_stack = [], [0] * n
        G.vis_cnt = 0
        G.sccs = []
        for i in range(n):
            if G.vis[i] == -1:
                G.dfs(i)
        invV = [-1] * n
        for i, scc in enumerate(G.sccs):
            for j in scc:
                invV[j] = i
        sccE = set()
        for i in range(n):
            for j in G[i]:
                sccE.add((invV[i], invV[j]))
        sccE = list(sccE)
        return G.sccs, sccE

    def dfs(self, i):
        G = self
        G.vis[i] = G.low[i] = G.vis_cnt
        G.vis_cnt += 1
        G.stack.append(i)
        G.in_stack[i] = 1
        for j in G[i]:
            if G.vis[j] == -1:
                G.dfs(j)
                G.low[i] = min(G.low[i], G.low[j])
            elif G.in_stack[j]:
                G.low[i] = min(G.low[i], G.vis[j])
        if G.low[i] == G.vis[i]:
            scc = []
            s = G.stack
            while s[-1] != i:
                scc.append(s.pop())
            scc.append(s.pop())
            for x in scc:
                G.in_stack[x] = 0
            G.sccs.append(scc)
        return



import functools
import weakref
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast
from typing_extensions import Protocol
from typing import Callable, TypeVar, Generic
import types
# Cached property both in library and in object

_T = TypeVar('_T')
F = TypeVar('F', bound=Callable[..., Any])


def copy_function(f: F, name: Optional[str] = None) -> F:
    """Based on http://stackoverflow.com/a/6528148/190597"""
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=name or f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return cast(F, g)


class implemented_at(Generic[F]):
    '''
    Decorator that ignores the wrapped function and uses the
    parameter function instead.
    
    Useful for implementing methods in separate files.
        Docs are copied so that any editor can provide help(method).
        Types signatures are copied so that type checkers work seamlessly.
    
    For usage details see help(implemented_at.usage)
    '''

    def __init__(self, target: F) -> None:
        self.target = target

    def __call__(self, wrapped: Callable[..., Any]) -> F:
        return copy_function(self.target, wrapped.__name__)

    @staticmethod
    def usage():
        '''
        # --- main.py ---
        (from .) import other_file

        class Hello:
            @implemented_at(other_file.hello)
            def hello(self):
                ...

        # --- other_file.py ---
        from __future__ import annotations
        from typing import TYPE_CHECKING, Type
        if TYPE_CHECKING:
            from (.)main import Hello

        def hello(obj: Hello, name:str):
            'I just say "hello {name}" for some given name'
            print(f'hello {name}')

        # --- interactive console ---
        h = Hello()
        help(h.hello)
        h.hello('world')
        '''


class _Property(Protocol[_T]):
    __get__: Callable[..., _T]


class _NormalMethod(Protocol[_T]):
    __name__: str
    __call__: Callable[..., _T]


def cached_method(maxsize=128, typed=False):
    '''decorator for converting a method into an lru cached method'''

    # https://stackoverflow.com/a/33672499/3671939
    def decorator(func: _NormalMethod[_T]) -> _NormalMethod[_T]:

        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            weak_self = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(maxsize=maxsize, typed=typed)
            def _cached_method(*args, **kwargs):
                self = weak_self()
                assert self
                return func(self, *args, **kwargs)

            setattr(self, func.__name__, _cached_method)
            return _cached_method(*args, **kwargs)

        return wrapped_func

    return decorator


class cached_property(property, _Property[_T]):
    '''
    decorator for converting a method into a cached property
    See https://stackoverflow.com/a/4037979/3671939
    This uses a modification:
    1. inherit from property, which disables setattr(instance, name, value)
        as it raises AttributeError: Can't set attribute
    2. use instance.__dict__[name] = value to fix
    '''

    def __init__(self, method: Callable[..., _T]):
        self._method = method

    def __get__(self, instance, _) -> _T:
        name = self._method.__name__
        value = self._method(instance)
        instance.__dict__[name] = value
        return value

    # def delete(self, instance, property_name: str):
    #     del instance.__dict__[property_name]

    @classmethod
    def set_property(cls, instance, property_name: str, value):
        instance.__dict__[property_name] = value

    @classmethod
    def is_computed(cls, instance, property_name: str):
        return property_name in instance.__dict__


class globally_cached_property(cached_property[_T]):
    '''
    decorator for converting a method into a cached property
    See https://stackoverflow.com/a/4037979/3671939
    This uses a modification:
    1. inherit from property, which disables setattr(instance, name, value)
        as it raises AttributeError: Can't set attribute
    2. use instance.__dict__[name] = value to fix
    '''
    cache = {}

    def __get__(self, instance, _) -> _T:
        name = self._method.__name__
        try:
            value = self.cache[(name, instance)]
        except KeyError:
            value = self._method(instance)
            self.cache[(name, instance)] = value
        instance.__dict__[name] = value
        return value

    def set_property(self, instance, property_name: str, value):
        instance.__dict__[property_name] = value
        self.cache[(property_name, instance)] = value



from ._random import (
    random_poset as poset,
    random_lattice as lattice,
    random_lattice_czech as lattice_czech,
    random_f_monotone_A as f_monotone_A,
    random_f_monotone_B as f_monotone_B,
    random_f_preserving_lub as f_preserving_lub,
)



from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, List, Tuple, Type, TypeVar, Union, cast
import numpy as np

from ..utils.algorithm_floyd_warshall import transitive_closure, floyd_warshall
from ..utils.numpy_types import npBoolMatrix, npUInt64Matrix


def parents_to_children(parents: List[List[int]]):
    n = len(parents)
    children = [[] for _ in range(n)]
    for ch in range(n):
        for pa in parents[ch]:
            children[pa].append(ch)
    return children


def child_to_dist(child: npBoolMatrix):
    'Compute all pairs shortest distances using Floyd-Warshall algorithm'
    # To do: use toposort or repeated dijsktra if assume_poset==True
    dist = floyd_warshall(child, infinity=child.shape[0])
    return dist


def children_to_leq(children: List[List[int]]):
    n = len(children)
    child = np.zeros((n, n), dtype=bool)
    for pa in range(n):
        for ch in children[pa]:
            child[ch, pa] = True
    dist = child_to_dist(child)
    leq = dist < n
    leq.flags.writeable = False
    dist.flags.writeable = False
    return leq, child, dist


def up_edges_to_leq(n, edges: Iterable[Tuple[int, int]]):
    'create Poset of size n respecting all given relations (descendant, ancestor)'
    leq = np.zeros((n, n), dtype=bool)
    leq[np.diag_indices_from(leq)] = True
    for des, anc in edges:
        leq[des, anc] = True
    leq = transitive_closure(leq)
    leq.flags.writeable = False
    return leq


from __future__ import annotations
import functools
from typing import Any, Callable, Generic, Iterable, List, Optional, Sequence, Set, Tuple, Type, TypeVar, Union, cast, overload
import numpy as np

from . import functional
from .utils.methodtools import cached_property, cached_method
from .utils.algorithm_tarjan import Tarjan
from .utils.numpy_types import npBoolMatrix, npUInt64Matrix
from . import utils
from .utils.algorithm_floyd_warshall import floyd_warshall, transitive_closure
from .utils.methodtools import implemented_at
from .base_methods import validation, identity, graphs, interface, description, graphviz

from . import base_methods as MD

from .base_methods.validation import PosetExceptions

Endomorphism = List[int]
PartialEndomorphism = Union[List[Optional[int]], List[None], Endomorphism]
_T = TypeVar('_T')


def _copy_as_type(self: Relation, cls: Type[_T], check=False) -> _T:
    kwargs = self.__dict__.copy()
    kwargs.pop('leq')
    kwargs.pop('labels', None)
    cpy = cls(self.leq, labels=self.labels, check=check, **kwargs)
    return cpy


def _lt_pairs(self: Relation):
    for i in range(self.n):
        for j in range(self.n):
            if self.leq[i, j] and i != j:
                yield i, j
    return


class Relation:
    '''
    Class for boolean relation matrices intended mostly for asserting that
    a matrix relation can be used with the fully featured parent classes:
    Poset and Lattice.

    The attribute name leq (less or equal) is missleading for arbitrary relations
    on purpose, because it is mostly used in parent classes.
    '''

    n: int
    leq: npBoolMatrix
    labels: Optional[Sequence[str]]

    MD = MD  # Provide access to library methods for all instances

    def __init__(self, leq: npBoolMatrix, check: bool,
                 labels: Optional[Sequence[str]] = None, **kwargs):
        self.n = len(leq)
        self.leq = leq
        self.labels = labels
        validation.validate_matrix_format(self)
        for k, v in kwargs.items():
            self.__dict__[k] = v
        if check:
            self.validate()
        return

    def validate(self):
        pass  # All binary matrices are valid relations

    @cached_property
    def is_poset(self):
        method = lambda: self.as_poset(check=True) and None
        return validation.ValidationError.capture(method)

    def as_poset(self, check: bool):
        return self.as_type(Poset, check=check)

    def as_type(self, cls: Type[_T], check=False) -> _T:
        if isinstance(self, cls):
            return self
        return _copy_as_type(self, cls, check=check)

    def try_type(self, cls: Type[_T]) -> Optional[_T]:
        try:
            return self.as_type(cls, check=True)
        except validation.ValidationError:
            return None

    def copy(self, check=False):
        return _copy_as_type(self, self.__class__, check=check)

    def reindex(self, rank: List[int], inverse=False):
        return identity.reindex(self, rank, inverse=inverse)

    @property
    def _labels(self):
        return self.labels or [f'{i}' for i in range(self.n)]

    def relabel(self, labels: Optional[Sequence[str]] = None):
        'copy of self with different labels'
        if labels == self.labels:
            return self
        if labels is not None:
            n = self.n
            m = len(labels)
            assert m == n, f'{m} labels found. Expected {n}'
            non = [l for l in labels if not isinstance(l, str)]
            assert not non, f'non-string label found: {non[0]}'
        Q = self.copy()
        Q.labels = labels
        return Q

    @classmethod
    def total(cls, n: int):
        'total order of n elements'
        G = [[i - 1] if i > 0 else [] for i in range(n)]
        return cls.from_children(G, check=False)

    '''
    @section
        Interface methods
    '''

    @classmethod
    def from_parents(cls, parents: List[List[int]], labels=None, check=True):
        'new instance from list: parents[i] = list of parents of i'
        children = interface.parents_to_children(parents)
        return cls.from_children(children, labels, check)

    @classmethod
    def from_children(cls, children: List[List[int]], labels=None, check=True):
        'new instance from list: children[i] = list of covers of i'
        leq, child, dist = interface.children_to_leq(children)
        leq.flags.writeable = False
        dist.flags.writeable = False
        return cls(leq, check=check, labels=labels, dist=dist, child=child,
                   children=children)

    @classmethod
    def from_down_edges(cls, n: int, edges: Iterable[Tuple[int, int]],
                        labels=None, check=True):
        'new instance of size n respecting all given relations (ancestor, descendant)'
        up_edges = [(j, i) for i, j in edges]
        return cls.from_up_edges(n, up_edges, labels, check)

    @classmethod
    def from_up_edges(cls, n: int, edges: Iterable[Tuple[int, int]],
                      labels=None, check=True):
        'new instance of size n respecting all given relations (descendant, ancestor)'
        leq = interface.up_edges_to_leq(n, edges)
        leq.flags.writeable = False
        return cls(leq, check=check, labels=labels)

    @classmethod
    def from_lambda(cls, elems: List[_T], f_leq: Callable[[_T, _T], bool],
                    labels=None, check=True):
        'new instance with: leq[i,j] = f_leq(elems[i], elems[j])'
        m = len(elems)
        leq = np.zeros((m, m), dtype=bool)
        for i in range(m):
            for j in range(m):
                leq[i, j] = f_leq(elems[i], elems[j])
        leq.flags.writeable = False
        return cls(leq, check=check, labels=labels)

    def _description(self):
        out = '\n'.join([
            f'Relation matrix:\n{self.leq.astype(int)}',
            f'Reflexive? {validation.is_reflexive(self),}',
            f'Antisymmetric? {validation.is_antisymmetric(self),}',
            f'Transitive? {validation.is_transitive(self),}',
        ])
        return out

    # def show(self, labels=None, save=None):
    #     'Display the relation using graphviz. Groups SCCs together'
    #     scc_components, scc_edges = self.scc_reduction()
    #     if labels is None:
    #         labels = [f'{i}' for i in range(self.n)]
    #     n = len(scc_components)
    #     labels = ['-'.join(labels[i] for i in I) for I in scc_components]
    #     return graphviz(n, edges=scc_edges, labels=labels, save=save)
    '''
    @section
        Graph operations
    '''

    def scc_reduction(self):
        n = self.n
        rel = self.leq
        G = [[] for _ in range(n)]
        for i in range(n):
            for j in range(n):
                if rel[i, j] and i != j:
                    G[i].append(j)
        return Tarjan(G).tarjan()

    def f_is_monotone(self, f: Endomorphism):
        for i, j in _lt_pairs(self):
            if not self.leq[f[i], f[j]]:
                return False
        return True


class Poset(Relation):
    '''
    Hashable object that represents an inmutable finite partial order.
    Uses a matrix and hashing is invariant under permutations.

    Run print(Poset.usage) for details and usage examples.

    The main attributes (always present) are:
        n: size of the poset.
            The elements of the poset are range(n)
        leq: read only less-or-equal boolean nxn matrix:
            leq[i,j] is True if and only "i <= j" in the poset order
        
        labels: optional sequence of n strings. Only used for displaying
        _idx: optional permutation of 1..n (not the identity).
    '''

    def validate(self):
        super().validate()
        validation.assert_is_poset(self)

    def as_lattice(self, check: bool):
        return self.as_type(Lattice, check=check)

    @cached_property
    def is_lattice(self):
        return self.try_type(Lattice) is not None

    '''
    @section
        Methods for identity
    '''

    def __hash__(self):
        return self.hash

    @cached_property
    def hash(self):
        'hash number for this poset'
        return identity.hasher(sorted(self.hash_elems))

    @cached_property
    def hash_elems(self):
        'hash for each element of the poset w.r.t. the poset order'
        return identity._hash_elems(self, rounds=2, salt=0)

    def __eq__(self, other: Poset):
        'Equality up to isomorphism, i.e. up to reindexing'
        return identity.find_isomorphism(self, other) is not None

    @cached_property
    def hash(self):
        'hash number for this poset'
        return identity.hasher(sorted(self.hash_elems))

    @cached_property
    def canonical(self):
        rank = identity.canonical_rank(self)
        P = self.reindex(rank)
        P.labels = None
        return P

    '''
    @section
        Fundamental properties
    '''

    @cached_property
    def child(self) -> npBoolMatrix:
        '''
        nxn boolean matrix: transitive reduction of the poset.
        child[i,j] == True iff j covers i (with no elements inbetween)
        '''
        child = graphs.transitive_reduction(self.leq)
        child.flags.writeable = False
        return child

    @cached_property
    def children(self) -> List[List[int]]:
        ''' top-down adjoint list (j in G[i] iff i covers j)'''
        n = self.n
        child = self.child
        return [[j for j in range(n) if child[j, i]] for i in range(n)]

    @cached_property
    def parents(self) -> List[List[int]]:
        '''bottom-up adjoint list (j in G[i] iff j covers i)'''
        n = self.n
        child = self.child
        return [[j for j in range(n) if child[i, j]] for i in range(n)]

    @cached_property
    def dist(self) -> npUInt64Matrix:
        '''
        Matrix of shortest distance from i upwards to j through parents
        n represents infinity.
        '''
        return interface.child_to_dist(self.child)

    '''
    @section
        Display methods
    '''

    @implemented_at(graphviz.show)
    def show(self):
        ...

    def __repr__(self):
        return self.name

    @cached_property
    def name(self):
        return description.name(self)

    @cached_property
    def upside_down(self):
        geq = self.leq.T
        Q = self.__class__(geq, check=False, labels=self.labels,
                           upside_down=self)
        return Q

    @cached_property
    def toposort_bottom_up(self):
        return graphs.toposort_bottom_up(self)

    @cached_property
    def bottoms(self):
        return graphs.bottoms(self)

    @cached_property
    def tops(self):
        return graphs.tops(self)


class Lattice(Poset):

    def validate(self):
        super().validate()
        validation.assert_is_lattice(self)

    @cached_property
    def is_distributive(self):
        try:
            validation.assert_is_distributive(self)
        except validation.ValidationError:
            return False
        return True

    @cached_property
    def is_modular(self):
        try:
            validation.assert_is_modular(self)
        except validation.ValidationError:
            return False
        return True

    @cached_property
    def lub(self):
        'matrix of i lub j, i.e. i join j'
        n = self.n
        leq = self.leq
        lub_id = {tuple(leq[i, :]): i for i in range(n)}
        lub = np.zeros((n, n), int)
        for i in range(n):
            for j in range(n):
                above = tuple(leq[i, :] & leq[j, :])
                if above not in lub_id:
                    raise PosetExceptions.LUB_Inconsistency(i, j)
                lub[i, j] = lub_id[above]
        lub.flags.writeable = False
        return lub

    @property
    def glb(self):
        return self.upside_down.lub

    @cached_property
    def bottom(self):
        'unique bottom element of the Poset. Throws if not present'
        return validation.expect_unique_bottom(self)

    @cached_property
    def top(self):
        'unique top element of the Poset. Throws if not present'
        return validation.expect_unique_top(self)

    @cached_property
    def irreducibles(self):
        n = self.n
        children = self.children
        return tuple([i for i in range(n) if len(children[i]) == 1])

    @cached_property
    def irreducible_descendants(self):
        '''Irreducibles below x for each x'''
        Rn = range(self.n)
        I = self.irreducibles
        leq = self.leq
        return tuple([tuple([i for i in I if leq[i, x]]) for x in Rn])

    @cached_property
    def irreducible_components(self: Lattice):
        '''
        Components of join irreducibles in toposort order and children
        lists for each component
        '''
        n = self.n
        if n <= 1:  # no join irreducibles at all
            return (0, [], [])
        irr = self.irreducibles
        sub = graphs.subgraph(self, irr)
        subcomps = graphs.independent_components(sub)
        m = len(subcomps)
        irrcomps = [[irr[j] for j in subcomps[i]] for i in range(m)]
        m_topo, m_children = zip(
            *(graphs._toposort_children(self, irrcomps[i]) for i in range(m)))
        m_topo = cast(Tuple[List[int]], m_topo)
        m_children = cast(Tuple[List[List[int]]], m_children)
        return m, m_topo, m_children

    def lub_of_many(self, elems: Iterable[int]) -> int:
        return functools.reduce(
            lambda i, j: self.lub[i, j],
            elems,
            self.bottom,
        )

    def glb_of_many(self, elems: Iterable[int]) -> int:
        return functools.reduce(
            lambda i, j: self.glb[i, j],
            elems,
            self.top,
        )

    @implemented_at(functional.f_lub)
    def f_lub(self, *args, **kwargs):
        ...

    @implemented_at(functional.f_glb)
    def f_glb(self, *args, **kwargs):
        ...

    @implemented_at(functional.f_iter)
    def f_iter(self, *args, **kwargs):
        ...
